#ifndef _HSML_LANG_
#define _HSML_LANG_

#include "main.h"

#define PATH_STRING_LEN 1023

#include <dlfcn.h>

char cache;
char cacherule;
char *pwd;
char *tmp;
void (*cache_output_add)(char *, char *);
void (*cache_print)(char *);
void (*cache_exec_add)(char *, char *);

char lang[];
char file[];

bool init(void);
char *get_path_from_env(void);
void show_config(void);
void print_begin(char *buf);
void print_end(char *buf);
void print_end_nl(char *buf);
void code_begin(char *buf);
void code_end(char *buf);
void codevar(char *buf, char *p);
void exec(char *file, char *buf, char *tmp);
void cache_exec(char *path, char *tmp);

#endif
